# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utilities for accessing the ShapeChecker from a wrapping `check_shapes` decorator.
"""
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Iterator, Optional

from .checker import S, ShapeChecker, TensorSpecLike
from .config import get_enable_check_shapes
from .error_contexts import ErrorContext, FunctionCallContext

_shape_checker: ContextVar[ShapeChecker] = ContextVar("_shape_checker")


@contextmanager
def set_shape_checker(checker: ShapeChecker) -> Iterator[None]:
    """
    Sets the current shape checker in the context.
    """
    token = _shape_checker.set(checker)
    try:
        yield
    finally:
        _shape_checker.reset(token)


def get_shape_checker() -> ShapeChecker:
    """
    Get the :class:`ShapeChecker` from the wrapping :func:`check_shapes` decorator.

    Behaviour is undefined if you call this from a function that is not directly wrapped in
    :func:`check_shapes` or :func:`inherit_check_shapes`.
    """
    return _shape_checker.get()


def check_shape(
    shaped: S, tensor_spec: TensorSpecLike, context: Optional[ErrorContext] = None
) -> S:
    """
    Raise an error if a tensor has the wrong shape.

    This uses the :class:`ShapeChecker` from the wrapping :func:`check_shapes` decorator. Behaviour
    is undefined if you call this from a function that is not directly wrapped in
    :func:`check_shapes` or :func:`inherit_check_shapes`.

    Example:

    .. literalinclude:: /examples/test_check_shapes_examples.py
       :start-after: [intermediate_results]
       :end-before: [intermediate_results]
       :dedent:

    :param shaped: The object whose shape to check.
    :param tensor_spec: Specification to check the tensor against. See: `Shape specification`_.
    :param context: Information about where ``shaped`` is coming from, for improved error
        messages.
    :returns: ``shaped``, for convenience.
    """
    if not get_enable_check_shapes():
        return shaped

    if context is None:
        context = FunctionCallContext(check_shape).precompute()

    return get_shape_checker().check_shape(shaped, tensor_spec, context)



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A library for annotating and checking the shapes of tensors.

The main entry point is :func:`check_shapes`.

Example:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [basic]
   :end-before: [basic]
   :dedent:


Speed, and interactions with `tf.function`
++++++++++++++++++++++++++++++++++++++++++

Shape checking has some performance impact. Shape checking can be disabled to help alleviate this.
Shape checking can be set to one of three different states:

* ``ENABLED``. Shapes are checked wherever they can be.
* ``EAGER_MODE_ONLY``. Shapes are not checked within anything wrapped in :func:`tf.function`.
* ``DISABLED``. Shapes are never checked.

The state can be set with :func:`set_enable_check_shapes`:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [disable__manual]
   :end-before: [disable__manual]
   :dedent:

Alternatively you can use :func:`disable_check_shapes` to disable shape checking in smaller scopes:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [disable__context_manager]
   :end-before: [disable__context_manager]
   :dedent:

Beware that any function declared while shape checking is disabled, will continue not to check
shapes, even if shape checking is otherwise enabled again.

The default state is ``EAGER_MODE_ONLY``; which is appropriate for smaller project, experiments, and
notebooks. Write and debug your code in eager mode, and add :func:`tf.function` when you believe
your code is correct and you want it to run fast. For larger project you probably want to modify
this setting. In particular you may want to enable all shape checks in your unit tests. If you use
`pytest <https://docs.pytest.org/>`_ you can do this by updating your root ``conftest.py`` with:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [pytest_fixture]
   :end-before: [pytest_fixture]
   :dedent:


If shape checking is set to ``ENABLED`` and your code is wrapped in :func:`tf.function` shape checks
are performed while tracing graphs, but *not* compiled into the actual graphs.  This is considered a
feature as that means that :func:`check_shapes` doesn't impact the execution speed of your functions
after they have been compiled.


Best-effort checking
++++++++++++++++++++

This library will perform shape checks on a best-effort basis. Many things can prevent this library
from being able to check shapes. For example:

* Unknown shapes. Sometimes the library is not able to determine the shape of an object, and thus
  cannot check that object. For example ``Optional`` arguments with value ``None`` cannot be
  checked, and compiled TensorFlow code can have variables with an unknown shape.

* Use of variable-rank dimensions (see below). In general we cannot infer the size of variable-rank
  dimensions if there are multiple variable-rank specifications within the same shape specification
  (e.g. ``cov: [m..., n...]``). This library will try to learn the size of these variable-rank
  dimensions from neighbouring shape specifications, but this is not always possible. Use of
  ``broadcast`` with variable-rank dimensions makes it even harder to infer these values.


Check specification
+++++++++++++++++++

The shapes to check are specified by the arguments to :func:`check_shapes`. Each argument is a
string of the format:

.. code-block:: bnf

   <argument specifier> ":" <shape specifier> ["if" <condition>] ["#" <note>]


Argument specification
----------------------

The ``<argument specifier>`` must start with either the name of an argument to the decorated
function, or the special name ``return``. The value ``return`` refers to the value returned by the
function.

The ``<argument specifier>`` can then be modified to refer to elements of the object in several
ways:

* Use ``.<name>`` to refer to an attribute of an object:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [argument_ref_attribute]
     :end-before: [argument_ref_attribute]
     :dedent:

* Use ``[<index>]`` to refer to a specific element of a sequence. This is particularly useful if
  your function returns a tuple of values:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [argument_ref_index]
     :end-before: [argument_ref_index]
     :dedent:

* Use ``[all]`` to select all elements of a collection:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [argument_ref_all]
     :end-before: [argument_ref_all]
     :dedent:

* Use ``.keys()`` to select all keys of a mapping:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [argument_ref_keys]
     :end-before: [argument_ref_keys]
     :dedent:

* Use ``.values()`` to select all values of a mapping:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [argument_ref_values]
     :end-before: [argument_ref_values]
     :dedent:

.. note::

   We do not support looking up a specific key or value in a  ``dict``.

If the argument, or any of the looked-up values, are ``None`` the check is skipped. This is useful
for optional values:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [argument_ref_optional]
   :end-before: [argument_ref_optional]
   :dedent:


Shape specification
-------------------

Shapes are specified by the syntax:


.. code-block:: bnf

   "[" <dimension specifier 1> "," <dimension specifer 2> "," ... "," <dimension specifier n> "]"

where ``<dimension specifier i>`` is one of:

* ``<integer>``, to require that dimension to have that exact size:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [dimension_spec_constant]
     :end-before: [dimension_spec_constant]
     :dedent:

* ``<name>``, to bind that dimension to a variable. Dimensions bound to the same variable must
  have the same size, though that size can be anything:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [dimension_spec_variable]
     :end-before: [dimension_spec_variable]
     :dedent:

* ``None`` or ``.`` to allow exactly one single dimension without constraints:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [dimension_spec_anonymous__none]
     :end-before: [dimension_spec_anonymous__none]
     :dedent:

  or:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [dimension_spec_anonymous__dot]
     :end-before: [dimension_spec_anonymous__dot]
     :dedent:

* ``*<name>`` or ``<name>...``, to bind *any* number of dimensions to a variable. Again,
  multiple uses of the same variable name must match the same dimension sizes:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [dimension_spec_variable_rank__star]
     :end-before: [dimension_spec_variable_rank__star]
     :dedent:

  or:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [dimension_spec_variable_rank__ellipsis]
     :end-before: [dimension_spec_variable_rank__ellipsis]
     :dedent:

* ``*`` or ``...``, to allow *any* number of dimensions without constraints:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [dimension_spec_anonymous_variable_rank__star]
     :end-before: [dimension_spec_anonymous_variable_rank__star]
     :dedent:

  or:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [dimension_spec_anonymous_variable_rank__ellipsis]
     :end-before: [dimension_spec_anonymous_variable_rank__ellipsis]
     :dedent:

A scalar shape is specified by ``[]``:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [dimension_spec__scalar]
   :end-before: [dimension_spec__scalar]
   :dedent:

Any of the above can be prefixed with the keyword ``broadcast`` to allow any value that broadcasts
to the specification. For example:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [dimension_spec_broadcast]
   :end-before: [dimension_spec_broadcast]
   :dedent:

Specifically, to mark a dimension as ``broadcast`` means:

* If the specification is that the dimension should have size ``n``, then the actual dimension must
  have value ``1`` or ``n``.
* If all leading dimension specifications are also marked ``broadcast``, then the actual shape is
  allowed to be shorter than the specification — the dimension is allowed to be missing.


Condition specification
-----------------------

You can use the optional ``if <condition>`` syntax to conditionally evaluate shapes. If an ``if
<condition>`` is used, the specification is only appplied if ``<condition>`` evaluates to ``True``.
This is useful if shapes depend on other input parameters. Valid conditions are:

* ``<argument specifier>``, with the same syntax and rules as above, except that constructions that
  evaluates to multiple elements are disallowed. Uses the ``bool`` built-in to convert the value of
  the argument to a ``bool``:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [bool_spec_argument_ref]
     :end-before: [bool_spec_argument_ref]
     :dedent:

* ``<argument specifier> is None``, and ``<argument specifier> is not None``, with the usual rules
  for an ``<argument specifier>``, to test whether an argument is, or is not, ``None``. We currently
  only allow tests against ``None``, not general Python equality tests:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [bool_spec_argument_ref_is_none]
     :end-before: [bool_spec_argument_ref_is_none]
     :dedent:

* ``<left> or <right>``, evaluates to ``True`` if any of ``<left>`` or ``<right>`` evaluates to
  ``True`` and to ``False`` otherwise:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [bool_spec_or]
     :end-before: [bool_spec_or]
     :dedent:

* ``<left> and <right>``, evaluates to ``False`` if any of ``<left>`` or ``<right>`` evaluates to
  ``False`` and to ``True`` otherwise:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [bool_spec_and]
     :end-before: [bool_spec_and]
     :dedent:

* ``not <right>``, evaluates to the opposite value of ``<right>``:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [bool_spec_not]
     :end-before: [bool_spec_not]
     :dedent:

* ``(<exp>)``, uses parenthesis to change operator precedence, as usual.

Conditions can be composed to apply different specs, depending on function arguments:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [bool_spec__composition]
   :end-before: [bool_spec__composition]
   :dedent:

.. note::

   All specifications with either no ``if`` syntax or a ``<condition>`` that evaluates to ``True``
   will be applied. It is possible for multiple specifications to apply to the same value.


Note specification
------------------

You can add notes to your specifications using a ``#`` followed by the note. These notes will be
appended to relevant error messages and appear in rewritten docstrings. You can add notes in two
places:

* On a single line by itself, to add a note to the entire function:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [note_spec__global]
     :end-before: [note_spec__global]
     :dedent:

* After the specification of a single argument, to add a note to that argument only:

  .. literalinclude:: /examples/test_check_shapes_examples.py
     :start-after: [note_spec__local]
     :end-before: [note_spec__local]
     :dedent:


Shape reuse
+++++++++++

Just like with other code it is useful to be able to specify a shape in one place and reuse the
specification. In particular this ensures that your code keep having internally consistent shapes,
even if it is refactored.


Class inheritance
-----------------

If you have a class hiererchy, you probably want to ensure that derived classes handle tensors with
the same shapes as the base classes. You can use the :func:`inherit_check_shapes` decorator to
inherit shapes from overridden methods:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [reuse__inherit_check_shapes]
   :end-before: [reuse__inherit_check_shapes]
   :dedent:


Functional programming
----------------------

If you prefer functional- over object oriented programming, you may have functions that you require
to handle the same shapes. To do this, remember that in Python a decorator is just a function, and
functions are objects that can be stored:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [reuse__functional]
   :end-before: [reuse__functional]
   :dedent:


Other reuse of shapes
---------------------

You can use :func:`get_check_shapes` to get, and reuse, the shape definitions from a previously
declared function. This is particularly useful to ensure fakes in tests use the same shapes as the
production implementation:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [reuse__get_check_shapes]
   :end-before: [reuse__get_check_shapes]
   :dedent:


Checking intermediate results
+++++++++++++++++++++++++++++

You can use the function :func:`check_shape` to check the shape of an intermediate result. This
function will use the same namespace as the immediately surrounding :func:`check_shapes` decorator,
and should only be called within functions that has such a decorator. For example:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [intermediate_results]
   :end-before: [intermediate_results]
   :dedent:


Checking shapes without a decorator
+++++++++++++++++++++++++++++++++++

While the :func:`check_shapes` decorator is the recommend way to use this library, it is possible to
use it without the decorator. In fact the decorator is just a wrapper around the class
:class:`ShapeChecker`, which can be used to check shapes directly:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [shape_checker__raw]
   :end-before: [shape_checker__raw]
   :dedent:

You can use the function :func:`get_shape_checker` to get the :class:`ShapeChecker` used by any
immediately surrounding :func:`check_shapes` decorator.


Documenting shapes
++++++++++++++++++

The :func:`check_shapes` decorator rewrites the docstring (``.__doc__``) of the decorated function
to add information about shapes, in a format compatible with
`Sphinx <https://www.sphinx-doc.org/en/master/>`_.

Only functions that already have a docstring will be updated. Functions that have no docstring at
all will not have one added, this is so that we do not override a docstring that would have been
inherited from a super class.

For example:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [doc_rewrite__definition]
   :end-before: [doc_rewrite__definition]
   :dedent:

will have ``.__doc__``:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [doc_rewrite__rewritten]
   :end-before: [doc_rewrite__rewritten]
   :dedent:

if you do not wish to have your docstrings rewritten, you can disable it with
:func:`set_rewrite_docstrings`:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [doc_rewrite__disable]
   :end-before: [doc_rewrite__disable]
   :dedent:


Supported types
+++++++++++++++

This library has built-in support for checking the shapes of:

* Python built-in scalars: ``bool``, ``int``, ``float`` and ``str``.
* Python built-in sequences: ``tuple`` and ``list``.
* NumPy ``ndarray``\\ s.
* TensorFlow ``Tensor``\\ s and ``Variable``\\ s.
* TensorFlow Probability ``DeferredTensor``\\ s, including ``TransformedVariable`` and
  :class:`gpflow.Parameter`.


Shapes of custom types
----------------------

:mod:`check_shapes` uses the function :func:`get_shape` to extract the shape of an object.
You can use :func:`register_get_shape` to extend :func:`get_shape` to extract shapes for you own
custom types:

.. literalinclude:: /examples/test_check_shapes_examples.py
   :start-after: [custom_type]
   :end-before: [custom_type]
   :dedent:
"""

from .accessors import get_check_shapes
from .base_types import Dimension, Shape
from .checker import ShapeChecker
from .checker_context import check_shape, get_shape_checker
from .config import (
    DocstringFormat,
    ShapeCheckingState,
    disable_check_shapes,
    get_enable_check_shapes,
    get_enable_function_call_precompute,
    get_rewrite_docstrings,
    set_enable_check_shapes,
    set_enable_function_call_precompute,
    set_rewrite_docstrings,
)
from .decorator import check_shapes
from .error_contexts import ErrorContext
from .inheritance import inherit_check_shapes
from .shapes import get_shape, register_get_shape

__version__ = "0.1.0"

__all__ = [
    "Dimension",
    "DocstringFormat",
    "ErrorContext",
    "Shape",
    "ShapeChecker",
    "ShapeCheckingState",
    "accessors",
    "argument_ref",
    "base_types",
    "bool_specs",
    "check_shape",
    "check_shapes",
    "checker",
    "checker_context",
    "config",
    "decorator",
    "disable_check_shapes",
    "error_contexts",
    "exceptions",
    "get_check_shapes",
    "get_enable_check_shapes",
    "get_enable_function_call_precompute",
    "get_rewrite_docstrings",
    "get_shape",
    "get_shape_checker",
    "inherit_check_shapes",
    "inheritance",
    "parser",
    "register_get_shape",
    "set_enable_check_shapes",
    "set_enable_function_call_precompute",
    "set_rewrite_docstrings",
    "shapes",
    "specs",
]



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Code for setting type hints, depending on library versions.
"""
import sys

from numpy import __version__ as np_version
from packaging.version import Version

NP_VERSION = Version(np_version)

NP_TYPE_CHECKING = False
"""
Whether to type-check numpy arrays at all. Defaults to False, because we don't know which
versions a client might use.
"""

GENERIC_NP_ARRAYS = (sys.version_info >= (3, 9)) and (NP_VERSION >= Version("1.22.0"))
"""
Whether to use generic numpy arrays. This is not applied at all, if type checking and not
`NP_TYPE_CHECKING`.
"""


def compute_mypy_flags() -> str:  # pragma: no cover
    # pylint: disable=import-outside-toplevel
    from mypy.version import __version__ as mypy_version

    MYPY_VERSION = Version(mypy_version)

    flags = []

    def set_always(variable: str, value: bool) -> None:
        if value:
            flags.append("--always-true")
        else:
            flags.append("--always-false")
        flags.append(variable)

    set_always("NP_TYPE_CHECKING", True)
    set_always("GENERIC_NP_ARRAYS", NP_VERSION >= Version("1.21.0"))
    if MYPY_VERSION >= Version("0.940"):
        flags.extend(["--enable-error-code", "ignore-without-code"])

    return " ".join(flags)



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Code for inheriting shape checks from a super class.
"""
import inspect
from typing import Callable, Optional, cast

from .accessors import maybe_get_check_shapes
from .base_types import C
from .config import get_enable_check_shapes
from .decorator import null_check_shapes


def inherit_check_shapes(func: C) -> C:
    """
    Decorator that inherits the :func:`check_shapes` decoration from any overridden method in a
    super-class.

    Example:

    .. literalinclude:: /examples/test_check_shapes_examples.py
       :start-after: [reuse__inherit_check_shapes]
       :end-before: [reuse__inherit_check_shapes]
       :dedent:

    See: `Class inheritance`_.
    """
    if not get_enable_check_shapes():
        return null_check_shapes(func)

    return cast(C, _InheritCheckShapes(func))


class _InheritCheckShapes:
    """
    Implementation of inherit_check_shapes.

    The ``__set_name__`` hack is to get access to the class the method was declared on.
    See: https://stackoverflow.com/a/54316392 .
    """

    def __init__(self, func: C) -> None:
        self._func = func

    def __set_name__(self, owner: type, name: str) -> None:
        overridden_check_shapes: Optional[Callable[[C], C]] = None
        for parent in inspect.getmro(owner)[1:]:
            overridden_method = getattr(parent, name, None)
            if overridden_method is None:
                continue
            overridden_check_shapes = maybe_get_check_shapes(overridden_method)
            if overridden_check_shapes is None:
                continue
            break

        assert overridden_check_shapes is not None, (
            f"@inherit_check_shapes did not find any overridden method of name '{name}'"
            f" on class '{owner.__name__}'."
        )

        self._func.class_name = owner.__name__  # type: ignore[attr-defined]
        wrapped = overridden_check_shapes(self._func)
        setattr(owner, name, wrapped)



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument

from abc import ABC
from dataclasses import replace
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
)

from lark.exceptions import UnexpectedInput
from lark.lark import Lark
from lark.lexer import PatternRE, PatternStr, Token
from lark.tree import Tree

from .argument_ref import (
    RESULT_TOKEN,
    AllElementsRef,
    ArgumentRef,
    AttributeArgumentRef,
    IndexArgumentRef,
    KeysRef,
    RootArgumentRef,
    ValuesRef,
)
from .bool_specs import (
    BoolTest,
    ParsedAndBoolSpec,
    ParsedArgumentRefBoolSpec,
    ParsedBoolSpec,
    ParsedNotBoolSpec,
    ParsedOrBoolSpec,
)
from .config import DocstringFormat, get_rewrite_docstrings
from .error_contexts import (
    ArgumentContext,
    ErrorContext,
    LarkUnexpectedInputContext,
    MultipleElementBoolContext,
    StackContext,
)
from .exceptions import CheckShapesError, DocstringParseError, SpecificationParseError
from .specs import (
    ParsedArgumentSpec,
    ParsedDimensionSpec,
    ParsedFunctionSpec,
    ParsedNoteSpec,
    ParsedShapeSpec,
    ParsedTensorSpec,
)

_VARIABLE_RANK_LEADING_TOKEN = "*"
_VARIABLE_RANK_TRAILING_TOKEN = "..."


def _tree_children(tree: Tree[Token]) -> Iterable[Tree[Token]]:
    """ Return all the children of `tree` that are trees themselves. """
    return (child for child in tree.children if isinstance(child, Tree))


def _token_children(tree: Tree[Token]) -> Iterable[str]:
    """ Return the values of all the children of `tree` that are tokens. """
    return (child.value for child in tree.children if isinstance(child, Token))


class _TreeVisitor(ABC):
    """
    Functionality for visiting the nodes of parse-trees.

    This differs from the classes built-in in Lark, in that it allows passing `*args` and
    `**kwargs`.

    Subclasses should add methods with the same name as Lark rules. Those methods should take the
    parse tree of the rule, followed by any other `*args` and `**kwargs` you want. They may return
    anything.
    """

    def visit(self, tree: Tree[Token], *args: Any, **kwargs: Any) -> Any:
        name = tree.data
        visit = getattr(self, name, None)
        assert visit, f"No method found with name {name}."
        return visit(tree, *args, **kwargs)


class _ParseSpec(_TreeVisitor):
    def __init__(self, source: str) -> None:
        self._source = source

    def argument_spec(self, tree: Tree[Token]) -> ParsedArgumentSpec:
        argument_ref, shape_spec, *other_specs = _tree_children(tree)
        argument = self.visit(argument_ref, False)
        shape = self.visit(shape_spec)
        condition = None
        note = None
        for other_spec in other_specs:
            other = self.visit(other_spec)
            if isinstance(other, ParsedBoolSpec):
                assert condition is None
                condition = other
            else:
                assert isinstance(other, ParsedNoteSpec)
                assert note is None
                note = other
        tensor = ParsedTensorSpec(shape, note)
        return ParsedArgumentSpec(argument, tensor, condition)

    def bool_spec_or(self, tree: Tree[Token]) -> ParsedBoolSpec:
        left, right = _tree_children(tree)
        return ParsedOrBoolSpec(self.visit(left), self.visit(right))

    def bool_spec_and(self, tree: Tree[Token]) -> ParsedBoolSpec:
        left, right = _tree_children(tree)
        return ParsedAndBoolSpec(self.visit(left), self.visit(right))

    def bool_spec_not(self, tree: Tree[Token]) -> ParsedBoolSpec:
        (right,) = _tree_children(tree)
        return ParsedNotBoolSpec(self.visit(right))

    def bool_spec_argument_ref_is_none(self, tree: Tree[Token]) -> ParsedBoolSpec:
        (argument_ref,) = _tree_children(tree)
        return ParsedArgumentRefBoolSpec(self.visit(argument_ref, True), BoolTest.IS_NONE)

    def bool_spec_argument_ref_is_not_none(self, tree: Tree[Token]) -> ParsedBoolSpec:
        (argument_ref,) = _tree_children(tree)
        return ParsedArgumentRefBoolSpec(self.visit(argument_ref, True), BoolTest.IS_NOT_NONE)

    def bool_spec_argument_ref(self, tree: Tree[Token]) -> ParsedBoolSpec:
        (argument_ref,) = _tree_children(tree)
        return ParsedArgumentRefBoolSpec(self.visit(argument_ref, True), BoolTest.BOOL)

    def argument_ref_root(self, tree: Tree[Token], is_for_bool_spec: bool) -> ArgumentRef:
        (token,) = _token_children(tree)
        return RootArgumentRef(token)

    def argument_ref_attribute(self, tree: Tree[Token], is_for_bool_spec: bool) -> ArgumentRef:
        (source,) = _tree_children(tree)
        (token,) = _token_children(tree)
        return AttributeArgumentRef(self.visit(source, is_for_bool_spec), token)

    def argument_ref_index(self, tree: Tree[Token], is_for_bool_spec: bool) -> ArgumentRef:
        (source,) = _tree_children(tree)
        (token,) = _token_children(tree)
        return IndexArgumentRef(self.visit(source, is_for_bool_spec), int(token))

    def argument_ref_all(self, tree: Tree[Token], is_for_bool_spec: bool) -> ArgumentRef:
        (source,) = _tree_children(tree)
        self._disallow_multiple_element_bool_spec(source, is_for_bool_spec)
        return AllElementsRef(self.visit(source, is_for_bool_spec))

    def argument_ref_keys(self, tree: Tree[Token], is_for_bool_spec: bool) -> ArgumentRef:
        (source,) = _tree_children(tree)
        self._disallow_multiple_element_bool_spec(source, is_for_bool_spec)
        return KeysRef(self.visit(source, is_for_bool_spec))

    def argument_ref_values(self, tree: Tree[Token], is_for_bool_spec: bool) -> ArgumentRef:
        (source,) = _tree_children(tree)
        self._disallow_multiple_element_bool_spec(source, is_for_bool_spec)
        return ValuesRef(self.visit(source, is_for_bool_spec))

    def _disallow_multiple_element_bool_spec(
        self, source: Tree[Token], is_for_bool_spec: bool
    ) -> None:
        if is_for_bool_spec:
            meta = source.meta
            raise SpecificationParseError(
                MultipleElementBoolContext(self._source, meta.end_line, meta.end_column)
            )

    def tensor_spec(self, tree: Tree[Token]) -> ParsedTensorSpec:
        shape_spec, *note_specs = _tree_children(tree)
        shape = self.visit(shape_spec)
        if note_specs:
            (note_spec,) = note_specs
            note = self.visit(note_spec)
        else:
            note = None
        return ParsedTensorSpec(shape, note)

    def shape_spec(self, tree: Tree[Token]) -> ParsedShapeSpec:
        (dimension_specs,) = _tree_children(tree)
        return ParsedShapeSpec(self.visit(dimension_specs))

    def dimension_specs(self, tree: Tree[Token]) -> Tuple[ParsedDimensionSpec, ...]:
        return tuple(
            self.visit(dimension_spec, i) for i, dimension_spec in enumerate(_tree_children(tree))
        )

    def dimension_spec_broadcast(self, tree: Tree[Token], i: int) -> ParsedDimensionSpec:
        (dimension_spec,) = _tree_children(tree)
        child = self.visit(dimension_spec, i)
        assert isinstance(child, ParsedDimensionSpec)
        return replace(child, broadcastable=True)

    def dimension_spec_constant(self, tree: Tree[Token], i: int) -> ParsedDimensionSpec:
        (token,) = _token_children(tree)
        return ParsedDimensionSpec(
            constant=int(token), variable_name=None, variable_rank=False, broadcastable=False
        )

    def dimension_spec_variable(self, tree: Tree[Token], i: int) -> ParsedDimensionSpec:
        (token,) = _token_children(tree)
        return ParsedDimensionSpec(
            constant=None, variable_name=token, variable_rank=False, broadcastable=False
        )

    def dimension_spec_anonymous(self, tree: Tree[Token], i: int) -> ParsedDimensionSpec:
        return ParsedDimensionSpec(
            constant=None, variable_name=None, variable_rank=False, broadcastable=False
        )

    def dimension_spec_variable_rank(self, tree: Tree[Token], i: int) -> ParsedDimensionSpec:
        (token1, token2) = _token_children(tree)
        if token1 == _VARIABLE_RANK_LEADING_TOKEN:
            variable_name = token2
        else:
            assert token2 == _VARIABLE_RANK_TRAILING_TOKEN
            variable_name = token1
        return ParsedDimensionSpec(
            constant=None, variable_name=variable_name, variable_rank=True, broadcastable=False
        )

    def dimension_spec_anonymous_variable_rank(
        self, tree: Tree[Token], i: int
    ) -> ParsedDimensionSpec:
        return ParsedDimensionSpec(
            constant=None, variable_name=None, variable_rank=True, broadcastable=False
        )

    def note_spec(self, tree: Tree[Token]) -> ParsedNoteSpec:
        _hash_token, *note_tokens = _token_children(tree)
        return ParsedNoteSpec(" ".join(token.strip() for token in note_tokens))


class _RewritedocString(_TreeVisitor):
    def __init__(self, source: str, function_spec: ParsedFunctionSpec) -> None:
        self._source = source
        self._spec_lines = self._argument_specs_to_sphinx(function_spec.arguments)
        self._notes = tuple(note.note for note in function_spec.notes)
        self._indent = self._guess_indent(source)

    def _argument_specs_to_sphinx(
        self,
        argument_specs: Collection[ParsedArgumentSpec],
    ) -> Mapping[str, Sequence[str]]:
        result: Dict[str, List[str]] = {}
        for spec in argument_specs:
            result.setdefault(spec.argument_ref.root_argument_name, []).append(
                self._argument_spec_to_sphinx(spec)
            )
        for lines in result.values():
            lines.sort()
        return result

    def _argument_spec_to_sphinx(self, argument_spec: ParsedArgumentSpec) -> str:
        tensor_spec = argument_spec.tensor
        shape_spec = tensor_spec.shape
        out = []
        out.append(f"* **{repr(argument_spec.argument_ref)}**")
        out.append(" has shape [")
        out.append(self._shape_spec_to_sphinx(shape_spec))
        out.append("]")

        if argument_spec.condition is not None:
            out.append(" if ")
            out.append(self._bool_spec_to_sphinx(argument_spec.condition, False))

        out.append(".")

        if tensor_spec.note is not None:
            note_spec = tensor_spec.note
            out.append(" ")
            out.append(note_spec.note)
        return "".join(out)

    def _bool_spec_to_sphinx(self, bool_spec: ParsedBoolSpec, paren_wrap: bool) -> str:
        if isinstance(bool_spec, ParsedOrBoolSpec):
            result = (
                self._bool_spec_to_sphinx(bool_spec.left, True)
                + " or "
                + self._bool_spec_to_sphinx(bool_spec.right, True)
            )
        elif isinstance(bool_spec, ParsedAndBoolSpec):
            result = (
                self._bool_spec_to_sphinx(bool_spec.left, True)
                + " and "
                + self._bool_spec_to_sphinx(bool_spec.right, True)
            )
        elif isinstance(bool_spec, ParsedNotBoolSpec):
            result = "not " + self._bool_spec_to_sphinx(bool_spec.right, True)
        else:
            assert isinstance(bool_spec, ParsedArgumentRefBoolSpec)
            if bool_spec.bool_test == BoolTest.BOOL:
                paren_wrap = False  # Never wrap a stand-alone argument.
                result = f"*{bool_spec.argument_ref!r}*"
            elif bool_spec.bool_test == BoolTest.IS_NONE:
                result = f"*{bool_spec.argument_ref!r}* is *None*"
            else:
                assert bool_spec.bool_test == BoolTest.IS_NOT_NONE
                result = f"*{bool_spec.argument_ref!r}* is not *None*"

        if paren_wrap:
            result = f"({result})"

        return result

    def _shape_spec_to_sphinx(self, shape_spec: ParsedShapeSpec) -> str:
        return ", ".join(self._dim_spec_to_sphinx(dim) for dim in shape_spec.dims)

    def _dim_spec_to_sphinx(self, dim_spec: ParsedDimensionSpec) -> str:
        tokens = []

        if dim_spec.broadcastable:
            tokens.append("broadcast ")

        if dim_spec.constant is not None:
            tokens.append(str(dim_spec.constant))
        elif dim_spec.variable_name:
            tokens.append(f"*{dim_spec.variable_name}*")
        else:
            if not dim_spec.variable_rank:
                tokens.append(".")

        if dim_spec.variable_rank:
            tokens.append("...")

        return "".join(tokens)

    def _guess_indent(self, docstring: str) -> Optional[int]:
        """
        Infer the level of indentation of a docstring.

        Returns `None` if the indentation could not be inferred.
        """
        # Algorithm adapted from:
        #     https://www.python.org/dev/peps/pep-0257/#handling-docstring-indentation

        # Convert tabs to spaces (following the normal Python rules)
        # and split into a list of lines:
        lines = docstring.expandtabs().splitlines()
        # Determine minimum indentation (first line doesn't count):
        no_indent = -1
        indent = no_indent
        for line in lines[1:]:
            stripped = line.lstrip()
            if not stripped:
                continue
            line_indent = len(line) - len(stripped)
            if indent == no_indent or line_indent < indent:
                indent = line_indent
        return indent if indent != no_indent else None

    def _insert_spec_lines(
        self, out: List[str], pos: int, spec_lines: Sequence[str], docs: Tree[Token]
    ) -> int:
        leading_str = self._source[pos : docs.meta.start_pos].rstrip()
        docs_start = pos + len(leading_str)
        docs_str = self._source[docs_start : docs.meta.end_pos]
        trailing_str = docs_str.lstrip()

        docs_indent = self._guess_indent(docs_str)
        if docs_indent is None:
            if self._indent is None:
                docs_indent = 4
            else:
                docs_indent = self._indent + 4
        indent_str = "\n" + docs_indent * " "

        out.append(leading_str)
        for spec_line in spec_lines:
            out.append(indent_str)
            out.append(spec_line)
        out.append("\n")
        out.append(indent_str)
        out.append(trailing_str)
        return docs.meta.end_pos

    def _insert_param_info_fields(
        self,
        is_first_info_field: bool,
        spec_lines: Mapping[str, Sequence[str]],
        out: List[str],
        pos: int,
    ) -> int:
        leading_str = self._source[pos:].rstrip()
        out.append(leading_str)
        pos += len(leading_str)

        if not self._source:
            # Case where nothing preceeds these fields. Just write them.
            needed_newlines = 0
        elif is_first_info_field:
            # Free-form documentation preceeds these fields. Have 2 newlines to separate them.
            needed_newlines = 2
        else:
            # Another info-field preceeds these fields.
            needed_newlines = 1

        indent = self._indent or 0
        indent_str = indent * " "
        indent_one_str = 4 * " "

        for arg_name, arg_lines in spec_lines.items():
            out.append(needed_newlines * "\n")
            needed_newlines = 1

            out.append(indent_str)
            if arg_name == RESULT_TOKEN:
                out.append(":returns:")
            else:
                out.append(f":param {arg_name}:")
            for arg_line in arg_lines:
                out.append("\n")
                out.append(indent_str)
                out.append(indent_one_str)
                out.append(arg_line)

        return pos

    def docstring(self, tree: Tree[Token]) -> str:
        # The strategy here is:
        # * `out` contains a list of strings that will be concatenated and form the final result.
        # * `pos` is the position such that `self._source[:pos]` has already been added to `out`,
        #   and `self._source[pos:]` still needs to be added.
        # * When visiting children we pass `out` and `pos`, and the children add content to `out`
        #   and return a new `pos`.
        docs, info_fields = _tree_children(tree)
        out: List[str] = []
        pos = 0

        if self._notes:
            if not docs.meta.empty:
                out.append(self._source[pos : docs.meta.end_pos])
                pos = docs.meta.end_pos
            indent = self._indent or 0
            indent_str = indent * " "
            for note in self._notes:
                if out:
                    out.append("\n\n")
                out.append(indent_str)
                out.append(note)

        pos = self.visit(info_fields, out, pos)
        out.append(self._source[pos:])

        return "".join(out)

    def info_fields(self, tree: Tree[Token], out: List[str], pos: int) -> int:
        spec_lines = dict(self._spec_lines)
        is_first_info_field = True
        for child in _tree_children(tree):
            # This will remove the self._spec_lines corresponding to found `:param:`'s.
            pos = self.visit(child, spec_lines, out, pos)
            is_first_info_field = False

        # Add any remaining `:param:`s:
        pos = self._insert_param_info_fields(is_first_info_field, spec_lines, out, pos)

        # Make sure info fields are terminated by a new-line:
        if self._spec_lines:
            if (pos >= len(self._source)) or (self._source[pos] != "\n"):
                out.append("\n")

        return pos

    def info_field_param(
        self, tree: Tree[Token], spec_lines: Dict[str, Sequence[str]], out: List[str], pos: int
    ) -> int:
        info_field_args, docs = _tree_children(tree)
        arg_name = self.visit(info_field_args)
        arg_lines = spec_lines.pop(arg_name, None)
        if arg_lines:
            pos = self._insert_spec_lines(out, pos, arg_lines, docs)
        return pos

    def info_field_returns(
        self, tree: Tree[Token], spec_lines: Dict[str, Sequence[str]], out: List[str], pos: int
    ) -> int:
        (docs,) = _tree_children(tree)
        return_lines = spec_lines.pop(RESULT_TOKEN, None)
        if return_lines:
            pos = self._insert_spec_lines(out, pos, return_lines, docs)
        return pos

    def info_field_other(
        self, tree: Tree[Token], spec_lines: Dict[str, Sequence[str]], out: List[str], pos: int
    ) -> int:
        return pos

    def info_field_args(self, tree: Tree[Token]) -> str:
        tokens = list(_token_children(tree))
        if not tokens:
            return ""
        return tokens[-1]


class _CachedParser:
    """
    Small wrapper around Lark so that we can reuse as much code as possible between the different
    things we parse.
    """

    def __init__(
        self,
        grammar_filename: str,
        start_symbol: str,
        parser_name: str,
        re_terminal_descriptions: Mapping[str, str],
        transformer_class: Type[_TreeVisitor],
        exception_class: Type[CheckShapesError],
    ) -> None:
        self._cache: Dict[Tuple[str, Tuple[Any, ...]], Any] = {}
        self._parser = Lark.open(
            grammar_filename,
            rel_to=__file__,
            propagate_positions=True,
            start=start_symbol,
            parser=parser_name,
        )
        self._terminal_descriptions = {}
        self._transformer_class = transformer_class
        self._exception_class = exception_class

        # Pre-compute nice terminal descriptions for our error messages:
        missing = set()
        unused = dict(re_terminal_descriptions)
        for terminal in self._parser.terminals:
            name = terminal.name
            pattern = terminal.pattern
            if isinstance(pattern, PatternStr):
                description = f'"{pattern.value}"'
            else:
                assert isinstance(pattern, PatternRE)
                unused_description = unused.pop(name, None)
                # If we enter this `if` then the parser is misconfigured, so we never get here, even
                # during tests.
                if unused_description is None:  # pragma: no cover
                    missing.add(name)
                    description = "ERROR"
                else:
                    description = f"{re_terminal_descriptions[name]} (re={pattern.value})"
            self._terminal_descriptions[name] = description
        assert not unused, f"Redundant terminal descriptions were provided: {sorted(unused)}"
        assert not missing, f"Some RE terminals did not have a description: {sorted(missing)}"

    def parse(self, text: str, transformer_args: Tuple[Any, ...], context: ErrorContext) -> Any:
        sentinel = object()
        cache_key = (text, transformer_args)
        result = self._cache.get(cache_key, sentinel)
        if result is sentinel:
            try:
                tree = self._parser.parse(text)
            except UnexpectedInput as ui:
                raise self._exception_class(
                    StackContext(
                        context, LarkUnexpectedInputContext(text, ui, self._terminal_descriptions)
                    )
                ) from ui

            try:
                result = self._transformer_class(*transformer_args).visit(tree)
            except CheckShapesError as cse:
                raise self._exception_class(StackContext(context, cse.context)) from cse

            self._cache[cache_key] = result
        return result


_TENSOR_SPEC_PARSER = _CachedParser(
    grammar_filename="check_shapes.lark",
    start_symbol="tensor_spec",
    parser_name="lalr",
    re_terminal_descriptions={
        "NOTE_TEXT": "note / comment text",
        "CNAME": "variable name",
        "INT": "integer",
        "WS": "whitespace",
    },
    transformer_class=_ParseSpec,
    exception_class=SpecificationParseError,
)
_ARGUMENT_SPEC_PARSER = _CachedParser(
    grammar_filename="check_shapes.lark",
    start_symbol="argument_or_note_spec",
    parser_name="lalr",
    re_terminal_descriptions={
        "NOTE_TEXT": "note / comment text",
        "CNAME": "variable name",
        "INT": "integer",
        "WS": "whitespace",
    },
    transformer_class=_ParseSpec,
    exception_class=SpecificationParseError,
)
_SPHINX_DOCSTRING_PARSER = _CachedParser(
    grammar_filename="docstring.lark",
    start_symbol="docstring",
    parser_name="earley",
    re_terminal_descriptions={
        "ANY": "any text",
        "CNAME": "variable name",
        "INFO_FIELD_OTHER": "Sphinx info field",
        "PARAM": "Sphinx parameter field",
        "RETURNS": "Sphinx `return` field",
        "WS": "whitespace",
    },
    transformer_class=_RewritedocString,
    exception_class=DocstringParseError,
)


def parse_tensor_spec(tensor_spec: str, context: ErrorContext) -> ParsedTensorSpec:

    """
    Parse a `check_shapes` tensor specification.
    """
    result = _TENSOR_SPEC_PARSER.parse(tensor_spec, (tensor_spec,), context)
    assert isinstance(result, ParsedTensorSpec)
    return result


def parse_function_spec(function_spec: Sequence[str], context: ErrorContext) -> ParsedFunctionSpec:
    """
    Parse all `check_shapes` argument or note specification for a single function.
    """
    arguments = []
    notes = []
    for i, spec in enumerate(function_spec):
        argument_context = StackContext(context, ArgumentContext(i))
        parsed_spec = _ARGUMENT_SPEC_PARSER.parse(spec, (spec,), argument_context)
        if isinstance(parsed_spec, ParsedArgumentSpec):
            arguments.append(parsed_spec)
        else:
            assert isinstance(parsed_spec, ParsedNoteSpec)
            notes.append(parsed_spec)
    return ParsedFunctionSpec(tuple(arguments), tuple(notes))


def parse_and_rewrite_docstring(
    docstring: Optional[str], function_spec: ParsedFunctionSpec, context: ErrorContext
) -> Optional[str]:
    """
    Rewrite `docstring` to include the shapes specified by the `argument_specs`.
    """
    if docstring is None:
        return None

    docstring_format = get_rewrite_docstrings()
    if docstring_format == DocstringFormat.NONE:
        return docstring

    assert docstring_format == DocstringFormat.SPHINX, (
        f"Current docstring format is {docstring_format}, but I don't know how to rewrite that."
        " See `check_shapes.config.set_rewrite_docstrings`."
    )
    result = _SPHINX_DOCSTRING_PARSER.parse(docstring, (docstring, function_spec), context)
    assert isinstance(result, str)
    return result



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=broad-except

"""
Infrastructure for describing the context of an error.

The :class:`MessageBuilder` is used to format / indent messages nicely.

The :class:`ErrorContext` is a reusable bit of information about where/why an error occurred that
can be written to a :class:`MessageBuilder`.

:class:`ErrorContext`\\ s can be composed using the :class:`StackContext` and
:class:`ParallelContext`.

This allows reusable error messages in a consistent format.
"""
import inspect
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from lark.exceptions import UnexpectedCharacters, UnexpectedEOF, UnexpectedInput

from .base_types import Shape
from .config import get_enable_function_call_precompute

if TYPE_CHECKING:  # pragma: no cover
    from .argument_ref import ArgumentRef
    from .bool_specs import ParsedBoolSpec
    from .specs import ParsedNoteSpec, ParsedShapeSpec, ParsedTensorSpec

_UNKNOWN_FILE = "<Unknown file>"
_UNKNOWN_LINE = "<Unknown line>"
_DISABLED_FILE_AND_LINE = (
    f"{_UNKNOWN_FILE}:{_UNKNOWN_LINE}"
    " (Disabled. Call check_shapes.set_enable_function_call_precompute(True) to"
    " see this. (Slow.))"
)
_NONE_SHAPE = "<Tensor is None or has unknown shape>"

_NO_VALUE = object()
"""
Sentinel to represent "no value" in places where ``None`` is a valid value.
"""


class MessageBuilder:
    """
    Utility for formatting nested text.
    """

    def __init__(self, indent_str: str = "") -> None:
        self._indent_str = indent_str
        self._lines: List[Union[str, Tuple[str, ...], "MessageBuilder"]] = []
        self._column_widths: List[int] = []

    def add_line(self, text: Any) -> None:
        """
        Write a line, indented at the current level.

        Input is converted to a `str` using `str(text)`.
        """
        text_str = str(text)
        self._lines.append(text_str)

    def add_columned_line(self, *texts: Any) -> None:
        """
        Write a line of several values, left-aligned within the current indentation level.

        Inputs are converted to `str`s using `str(text)`.
        """
        widths = self._column_widths
        text_strs = tuple(str(t) for t in texts)
        for i, t in enumerate(text_strs):
            w = len(t)
            if i < len(widths):
                widths[i] = max(widths[i], w)
            else:
                widths.append(w)
        self._lines.append(text_strs)

    @contextmanager
    def indent(self) -> Iterator["MessageBuilder"]:
        """
        Context manager for indenting text.
        """
        child = MessageBuilder(self._indent_str + "  ")
        self._lines.append(child)
        yield child

    def build(self) -> str:
        """
        Compile all collected text into a single string.
        """
        fragments: List[str] = []

        def _collect(builder: "MessageBuilder") -> None:
            # pylint: disable=protected-access
            for line in builder._lines:
                if isinstance(line, str):
                    fragments.extend((builder._indent_str, line, "\n"))
                elif isinstance(line, tuple):
                    fragments.append(builder._indent_str)
                    for i, t in enumerate(line):
                        fragments.append(t)
                        if i + 1 < len(line):
                            fragments.append(" " * (builder._column_widths[i] - len(t) + 1))
                    fragments.append("\n")
                else:
                    assert isinstance(line, MessageBuilder)
                    _collect(line)

        _collect(self)
        return "".join(fragments)


class ErrorContext(ABC):
    """
    A context in which an error can occur.

    Contexts should be immutable, and implement :func:`__eq__` - so that they can be composed using
    :class:`StackContext` and :class:`ParallelContext`.

    The contexts are often created even if an error doesn't actually occur, so they should be cheap
    to create - prefer to do any slow computation in :func:`print`, rather than in :func:`__init__`.

    Maybe think of an :class:`ErrorContext` as a factory of error messages.
    """

    @abstractmethod
    def print(self, builder: MessageBuilder) -> None:
        """
        Print this context to the given `MessageBuilder`.
        """


@dataclass(frozen=True)
class StackContext(ErrorContext):
    """
    Error context where one context is "inside" another one.
    """

    parent: ErrorContext
    child: ErrorContext

    def print(self, builder: MessageBuilder) -> None:
        def _print(context: ErrorContext) -> None:
            nonlocal builder

            if isinstance(context, StackContext):
                _print(context.parent)
                _print(context.child)
            else:
                context.print(builder)

                with builder.indent() as b:
                    builder = b

        _print(self)


@dataclass(frozen=True)
class ParallelContext(ErrorContext):
    """
    Error context with many contexts in parallel.
    """

    children: Tuple[ErrorContext, ...]

    def print(self, builder: MessageBuilder) -> None:
        # Merge any nested parallel contexts.
        flat: List[ErrorContext] = []

        def _flatten(context: ErrorContext) -> None:
            if isinstance(context, ParallelContext):
                for child in context.children:
                    _flatten(child)
            else:
                flat.append(context)

        _flatten(self)

        # If any `StackContexts`s have the same (grand-) parent, merge them, so that we don't repeat
        # stuff in the output.
        def _split_head(context: StackContext) -> Tuple[ErrorContext, ErrorContext]:
            parent = context.parent
            child = context.child
            if isinstance(parent, StackContext):
                # Transform:
                #   /\         /\
                #  /\ C  -->  A /\
                # A  B         B  C
                # So we get:                     /\
                #             head = A,  body = B  C
                return _split_head(StackContext(parent.parent, StackContext(parent.child, child)))
            else:
                return parent, child

        by_head: Dict[ErrorContext, List[ErrorContext]] = {}
        for child in flat:
            if isinstance(child, StackContext):
                split_child = _split_head(child)
                head: ErrorContext = split_child[0]
                body: Optional[ErrorContext] = split_child[1]
            else:
                head = child
                body = None
            bodies = by_head.setdefault(head, [])
            if body is not None:
                bodies.append(body)

        for head, bodies in by_head.items():
            head.print(builder)
            with builder.indent() as b:
                ParallelContext(tuple(bodies)).print(b)


@dataclass(frozen=True)
class FunctionCallContext(ErrorContext):
    """
    An error occured inside a function that was called.

    Normally `print` should be called from within the called function: `func`.
    If that is impossible, first call `precompute` from within `func`.
    """

    func: Callable[..., Any]
    path_and_line: Optional[str] = None

    def _get_target(self) -> Optional[Tuple[str, str]]:
        func = inspect.unwrap(self.func)
        search_name: str = func.__name__
        search_path = inspect.getsourcefile(func)
        if search_path is None:
            return None
        return search_name, search_path

    def _count_wrappers(self) -> int:
        wrappers = -1
        f: Optional[Callable[..., Any]] = self.func
        while f is not None:
            wrappers += 1
            f = getattr(f, "__wrapped__", None)
        return wrappers

    def _get_calling_frame(self) -> Optional[inspect.FrameInfo]:
        search_name_path = self._get_target()
        if search_name_path is None:
            return None

        stack = inspect.stack(0)
        for i, frame in enumerate(stack):
            name_path = (frame.function, frame.filename)
            if search_name_path == name_path:
                return stack[i + self._count_wrappers() + 1]

        return None

    def _compute(self) -> str:
        if self.path_and_line is not None:
            return self.path_and_line

        frame = self._get_calling_frame()
        if frame is None:
            path = _UNKNOWN_FILE
            line = _UNKNOWN_LINE
        else:
            path = frame.filename
            line = str(frame.lineno)

        return f"{path}:{line}"

    def print(self, builder: MessageBuilder) -> None:
        builder.add_columned_line(f"{self.func.__name__} called at:", self._compute())

    def precompute(self) -> "FunctionCallContext":
        """
        Precompute the values to print.

        This is useful to capture the position of the stack of the relevant call, if this object is
        saved for later use.

        :returns: A new instance with precomptued values.
        """
        if get_enable_function_call_precompute():
            return FunctionCallContext(self.func, self._compute())
        else:
            return FunctionCallContext(self.func, _DISABLED_FILE_AND_LINE)


@dataclass(frozen=True)
class FunctionDefinitionContext(ErrorContext):
    """
    An error occured in the context of a function definition.
    """

    func: Callable[..., Any]

    def print(self, builder: MessageBuilder) -> None:
        name = self.func.__qualname__
        try:
            path = inspect.getsourcefile(self.func)
        except Exception:  # pragma: no cover
            path = _UNKNOWN_FILE
        try:
            _, line_int = inspect.getsourcelines(self.func)
            line = str(line_int)
        except Exception:  # pragma: no cover
            line = _UNKNOWN_LINE
        path_and_line = f"{path}:{line}"

        builder.add_columned_line("Function:", name)
        with builder.indent() as b:
            b.add_columned_line("Declared:", path_and_line)


@dataclass(frozen=True)
class VariableContext(ErrorContext):
    """
    An error occurred in the context of a shape specification variable.
    """

    variable_name: str

    def print(self, builder: MessageBuilder) -> None:
        builder.add_columned_line("Variable:", self.variable_name)


@dataclass(frozen=True)
class TensorSpecContext(ErrorContext):
    """
    An error occurred in the context of a tensor specification.
    """

    spec: "ParsedTensorSpec"

    def print(self, builder: MessageBuilder) -> None:
        builder.add_columned_line("Specification:", self.spec)


@dataclass(frozen=True)
class ArgumentContext(ErrorContext):
    """
    An error occurred in the context of an argument to a function.
    """

    name_or_index: Union[str, int]
    value: Any = _NO_VALUE

    def print(self, builder: MessageBuilder) -> None:
        if isinstance(self.name_or_index, str):
            builder.add_columned_line("Argument:", self.name_or_index)
        else:
            builder.add_columned_line("Argument number (0-indexed):", self.name_or_index)
        if self.value is not _NO_VALUE:
            with builder.indent() as b:
                b.add_columned_line("Value:", self.value)

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)
            and self.name_or_index == other.name_or_index
            and self.value is other.value
        )

    def __hash__(self) -> int:
        return hash((self.name_or_index, id(self.value)))


@dataclass(frozen=True)
class AttributeContext(ErrorContext):
    """
    An error occurred in the context of an attribute on an object.
    """

    name: str
    value: Any = _NO_VALUE

    def print(self, builder: MessageBuilder) -> None:
        builder.add_columned_line("Attribute:", f".{self.name}")
        if self.value is not _NO_VALUE:
            with builder.indent() as b:
                b.add_columned_line("Value:", self.value)

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.name == other.name and self.value is other.value

    def __hash__(self) -> int:
        return hash((self.name, id(self.value)))


@dataclass(frozen=True)
class IndexContext(ErrorContext):
    """
    An error occurred in the context of an index in a sequence.
    """

    index: int
    value: Any = _NO_VALUE

    def print(self, builder: MessageBuilder) -> None:
        builder.add_columned_line("Index:", f"[{self.index}]")
        if self.value is not _NO_VALUE:
            with builder.indent() as b:
                b.add_columned_line("Value:", self.value)

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.index == other.index and self.value is other.value

    def __hash__(self) -> int:
        return hash((self.index, id(self.value)))


@dataclass(frozen=True)
class MappingKeyContext(ErrorContext):
    """
    An error occurent in the context of a key in a map.
    """

    key: Any

    def print(self, builder: MessageBuilder) -> None:
        builder.add_columned_line("Map key:", f"[{self.key!r}]")

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.key is other.key

    def __hash__(self) -> int:
        return id(self.key)


@dataclass(frozen=True)
class MappingValueContext(ErrorContext):
    """
    An error occurent in the context of a value in a map.
    """

    key: Any
    value: Any = _NO_VALUE

    def print(self, builder: MessageBuilder) -> None:
        builder.add_columned_line("Map value, of key:", f"[{self.key!r}]")
        if self.value is not _NO_VALUE:
            with builder.indent() as b:
                b.add_columned_line("Value:", self.value)

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.key is other.key and self.value is other.value

    def __hash__(self) -> int:
        return hash((id(self.key), id(self.value)))


@dataclass(frozen=True)
class ConditionContext(ErrorContext):
    """
    An error occurred in a conditional context.
    """

    condition: "ParsedBoolSpec"

    def print(self, builder: MessageBuilder) -> None:
        builder.add_columned_line("Condition:", self.condition)


@dataclass(frozen=True)
class ShapeContext(ErrorContext):
    """
    An error occurred in the context of the shapes of function arguments.
    """

    expected: "ParsedShapeSpec"
    actual: Shape

    def print(self, builder: MessageBuilder) -> None:
        if self.actual is None:
            actual_str = _NONE_SHAPE
        else:
            actual_str = f"[{', '.join(str(dim) for dim in self.actual)}]"
        builder.add_columned_line("Expected:", self.expected)
        builder.add_columned_line("Actual:", actual_str)


@dataclass(frozen=True)
class NoteContext(ErrorContext):
    """
    An error occurred in a context where a user has added a note.
    """

    note: "ParsedNoteSpec"

    def print(self, builder: MessageBuilder) -> None:
        builder.add_columned_line("Note:", self.note.note)


@dataclass(frozen=True)
class ObjectValueContext(ErrorContext):
    """
    An error was caused by the value of an object.
    """

    obj: Any

    def print(self, builder: MessageBuilder) -> None:
        builder.add_columned_line("Value:", repr(self.obj))

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.obj is other.obj

    def __hash__(self) -> int:
        return id(self.obj)


@dataclass(frozen=True)
class ObjectTypeContext(ErrorContext):
    """
    An error was caused by the type of an object.
    """

    obj: Any

    def print(self, builder: MessageBuilder) -> None:
        t = type(self.obj)
        builder.add_columned_line("Object type:", f"{t.__module__}.{t.__qualname__}")

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.obj is other.obj

    def __hash__(self) -> int:
        return id(self.obj)


@dataclass(frozen=True)  # type: ignore[misc]  # mypy doesn't like abstract `dataclasses`.
class ParserInputContext(ErrorContext, ABC):
    """
    Abstract base class for contexts that relate to parser input.
    """

    text: str

    def print_line(self, builder: MessageBuilder, line: int, column: int) -> None:
        if line > 0:
            line_content = self.text.split("\n")[line - 1]
            builder.add_columned_line("Line:", f'"{line_content}"')
            if column > 0:
                builder.add_columned_line("", " " * column + "^")


@dataclass(frozen=True)
class LarkUnexpectedInputContext(ParserInputContext):
    """
    An error was caused by an `UnexpectedInput` error from `Lark`.
    """

    error: UnexpectedInput
    terminal_descriptions: Mapping[str, str]

    def print(self, builder: MessageBuilder) -> None:
        self.print_line(
            builder,
            getattr(self.error, "line", -1),
            getattr(self.error, "column", -1),
        )

        expected: Sequence[str] = getattr(self.error, "accepts", [])
        if not expected:
            expected = getattr(self.error, "expected", [])
        expected = sorted(expected)
        expected_key = "Expected:" if len(expected) <= 1 else "Expected one of:"
        for expected_name in expected:
            expected_value = self.terminal_descriptions.get(expected_name, expected_name)
            builder.add_columned_line(expected_key, expected_value)
            expected_key = ""

        if isinstance(self.error, UnexpectedCharacters):
            builder.add_line("Found unexpected character.")
        if isinstance(self.error, UnexpectedEOF):
            builder.add_line("Found unexpected end of input.")

    def __hash__(self) -> int:
        return hash((self.error, *sorted(self.terminal_descriptions.items())))


@dataclass(frozen=True)
class MultipleElementBoolContext(ParserInputContext):
    """
    An error was caused by trying to use a multi-element argument specification as a bool.
    """

    line: int
    column: int

    def print(self, builder: MessageBuilder) -> None:
        self.print_line(builder, self.line, self.column)
        builder.add_line(
            "Argument references that evaluate to multiple values are not supported for boolean"
            " expressions."
        )



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class responsible for remembering and checking shapes.
"""
from dataclasses import dataclass, field
from typing import (
    Any,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from .base_types import Dimension, Shape
from .config import get_enable_check_shapes
from .error_contexts import (
    ErrorContext,
    FunctionCallContext,
    IndexContext,
    NoteContext,
    ParallelContext,
    ShapeContext,
    StackContext,
    TensorSpecContext,
    VariableContext,
)
from .exceptions import ShapeMismatchError, VariableTypeError
from .parser import parse_tensor_spec
from .shapes import get_shape
from .specs import ParsedDimensionSpec, ParsedShapeSpec, ParsedTensorSpec

S = TypeVar("S")


TensorSpecLike = Union[ParsedTensorSpec, str, Shape]
_Shape = Tuple[Dimension, ...]


def _as_parsed_tensor_spec(tensor_spec: TensorSpecLike, context: ErrorContext) -> ParsedTensorSpec:
    if isinstance(tensor_spec, ParsedTensorSpec):
        return tensor_spec
    elif isinstance(tensor_spec, str):
        return parse_tensor_spec(tensor_spec, context)
    else:
        dimension_specs = []
        if isinstance(tensor_spec, tuple):
            for dim in tensor_spec:
                if isinstance(dim, int):
                    dimension_specs.append(ParsedDimensionSpec(dim, None, False, False))
                else:
                    assert dim is None
                    dimension_specs.append(ParsedDimensionSpec(None, None, False, False))
        else:
            assert tensor_spec is None
            dimension_specs.append(ParsedDimensionSpec(None, None, True, False))
        shape = ParsedShapeSpec(tuple(dimension_specs))
        return ParsedTensorSpec(shape, None)


@dataclass
class _ObservedDim:
    """
    Storage of observed size of a single dimension variable.
    """

    size: Optional[int] = None

    def check_and_update(self, actual: Optional[int], broadcast: bool) -> bool:
        """
        Attempt to merge new data into this observation.

        Returns whether the new data is compatible with existing observations. If this method
        returns `False` this object may have been left in an invalid state and should not be used
        again.
        """
        if (actual is None) or (broadcast and actual == 1):
            # Update contains no information. Nothing to do.
            return True

        if self.size is None:
            self.size = actual

        return self.size == actual


@dataclass
class _ObservedDims:
    """
    Storage of observed sizes of a var-rank / batch variable.
    """

    sizes: Optional[List[_ObservedDim]] = None
    known_rank: bool = False

    def check_and_update(
        self,
        actual: Optional[Tuple[Optional[int], ...]],
        broadcast: bool,
        shape_possibly_truncated: bool,
    ) -> bool:
        """
        Attempt to merge new data into this observation.

        Returns whether the new data is compatible with existing observations. If this method
        returns `False` this object may have been left in an invalid state and should not be used
        again.
        """
        if actual is None:
            # Update contains no information. Nothing to do.
            return True

        if self.sizes is None:
            self.sizes = []
            assert not self.known_rank

        # First make sure lengths are set up and matches.
        longer = len(self.sizes) - len(actual)
        if self.known_rank:
            if shape_possibly_truncated:
                if longer < 0:
                    return False
            else:
                if longer != 0:
                    return False
        else:
            if longer < 0:
                self.sizes = [_ObservedDim() for _ in range(-longer)] + self.sizes
                longer = 0
            if shape_possibly_truncated:
                pass  # We don't know anything about total rank.
            else:
                if longer > 0:
                    return False
                self.known_rank = True
        assert longer >= 0

        # Then match individual dimensions.
        for i, actual_dim in enumerate(actual):
            if not self.sizes[i + longer].check_and_update(actual_dim, broadcast):
                return False

        return True


@dataclass(eq=False)
class _ShapeCheck:
    """
    A shape check that is waiting to be performed.
    """

    actual: _Shape
    """
    Actual observed shape.

    Only `actual[actual_begin:actual_end]` is still waiting to be checked. The beginning and end may
    already have been checked.
    """

    actual_begin: int
    actual_end: int

    expected: ParsedShapeSpec
    """
    Specification to check against.

    Only `expected[expected_begin:expected_end]` is still waiting to be checked. The beginning and
    end may already have been checked.
    """

    expected_begin: int
    expected_end: int

    @property
    def finished(self) -> bool:
        """
        Whether this entire check has been performed.
        """
        return self.expected_begin >= self.expected_end


@dataclass
class _VariableState:
    """
    Structure of stuff we need to know about each variable.
    """

    uses: List[Tuple[ParsedTensorSpec, ErrorContext]] = field(default_factory=list)
    """ List of specs where this variable is used. """

    observed_dim: Optional[_ObservedDim] = None
    """
    Observed size of this variable, if the variable is rank-1.

    Set this `None` if this variable is varrank.
    """

    observed_dims: Optional[_ObservedDims] = None
    """
    Observed shape of this variable, if the variable is varrank.

    Set this `None` if this variable is rank-1.
    """

    waiting_for_varrank: Set[_ShapeCheck] = field(default_factory=set)
    """ Checks that are waiting for the rank of this variable to be determined. """


class ShapeChecker:
    """
    Mechanism for checking the shapes of tensors.

    This remembers observed shapes and specifications, so that tensors can be checked for
    compatibility across multiple calls, and so that we can provide good error messages.

    Example:

    .. literalinclude:: /examples/test_check_shapes_examples.py
       :start-after: [shape_checker__raw]
       :end-before: [shape_checker__raw]
       :dedent:
    """

    def __init__(self) -> None:
        self._variables: DefaultDict[str, _VariableState] = DefaultDict(_VariableState)
        self._observed_shapes: List[Tuple[Shape, ParsedTensorSpec, ErrorContext]] = []
        self._additional_context: List[ErrorContext] = []

    def add_context(self, context: ErrorContext) -> None:
        """
        Add arbirtary context to the shape checker.

        This context will be included in any error messages.

        :param context: Context to add to this shape checker.
        """
        self._additional_context.append(context)

    def check_shape(
        self, shaped: S, tensor_spec: TensorSpecLike, context: Optional[ErrorContext] = None
    ) -> S:
        """
        Raise an error if a tensor has the wrong shape.

        This remembers observed shapes and specifications, so that tensors can be checked for
        compatibility across multiple calls, and so that we can provide good error messages.

        :param shaped: The object whose shape to check.
        :param tensor_spec: Specification to check the tensor against.
            Usually this is a ``str`` in the format described in `Shape specification`_.
            Alternatively this can be a pre-parsed :class:`ParsedTensorSpec`, or an actual
            :class:`Shape`.
        :param context: Information about where ``shaped`` is coming from, for improved error
            messages.
        :returns: ``shaped``, for convenience.
        """
        if not get_enable_check_shapes():
            return shaped

        if context is None:
            context = FunctionCallContext(self.check_shape).precompute()

        self.check_shapes([(shaped, tensor_spec, context)])

        return shaped

    def check_shapes(
        self,
        checks: Iterable[
            Union[Tuple[Any, TensorSpecLike], Tuple[Any, TensorSpecLike, ErrorContext]]
        ],
    ) -> None:
        """
        Raise an error if any tensor has the wrong shape.

        This remembers observed shapes and specifications, so that tensors can be checked for
        compatibility across multiple calls, and so that we can provide good error messages.

        :param checks: Checks to perform. The elements can either be ``(shaped, tensor_spec)`` or
            ``(shaped, tensor_spec, context)`` tuples. Where: ``shaped`` is the tensor whose shape
            to check; ``tensor_spec`` is the specification to check it against (see `Shape
            specification`_); and ``context`` contains (optional) information about where ``shaped``
            came from - for better error messages.
        """
        if not get_enable_check_shapes():
            return

        shape_check_queue = self._parse_checks(checks)
        while shape_check_queue:
            shape_check = shape_check_queue.pop()

            # Note that self._match_dims may not be able to match all dims. If that is the case it
            # adds the check to `_VariableState.waiting_for_varrank`:
            dim_checks = self._match_dims(shape_check)
            for dim_spec, actual_dims, shape_possibly_truncated in dim_checks:
                # Note that self._check_dim may revive checks from
                # `_VariableState.waiting_for_varrank` and add them back to `shape_check_queue`:
                self._check_dim(dim_spec, actual_dims, shape_possibly_truncated, shape_check_queue)

    def _parse_checks(
        self,
        checks: Iterable[
            Union[Tuple[Any, TensorSpecLike], Tuple[Any, TensorSpecLike, ErrorContext]]
        ],
    ) -> List[_ShapeCheck]:
        """
        Sanity check, register and parse the given ``checks`` into :class:`_ShapeCheck` objects.
        """
        shape_checks = []
        variable_type_errors = []
        call_context: Optional[ErrorContext] = None
        for i, check in enumerate(checks):
            shaped, tensor_spec, *contexts = check

            # Determine error context:
            if contexts:
                (context,) = contexts
            else:
                if call_context is None:
                    call_context = FunctionCallContext(self.check_shapes).precompute()
                context = StackContext(call_context, IndexContext(i))

            parsed_tensor_check = _as_parsed_tensor_spec(tensor_spec, context)
            parsed_shape_check = parsed_tensor_check.shape

            # Determine shape:
            if shaped is None:
                shape = None
            else:
                shape = get_shape(shaped, context)
            self._observed_shapes.append((shape, parsed_tensor_check, context))

            if shape is not None:
                shape_checks.append(
                    _ShapeCheck(
                        actual=shape,
                        actual_begin=0,
                        actual_end=len(shape),
                        expected=parsed_shape_check,
                        expected_begin=0,
                        expected_end=len(parsed_shape_check.dims),
                    )
                )

            # Record used variables:
            for dim_spec in parsed_shape_check.dims:
                if dim_spec.variable_name is None:
                    continue
                variable = self._variables[dim_spec.variable_name]
                variable.uses.append((parsed_tensor_check, context))
                if dim_spec.variable_rank:
                    if variable.observed_dims is None:
                        variable.observed_dims = _ObservedDims()
                        if variable.observed_dim is not None:
                            variable_type_errors.append(dim_spec.variable_name)
                else:
                    if variable.observed_dim is None:
                        variable.observed_dim = _ObservedDim()
                        if variable.observed_dims is not None:
                            variable_type_errors.append(dim_spec.variable_name)

        if variable_type_errors:
            error_contexts = [
                StackContext(
                    VariableContext(variable),
                    ParallelContext(
                        tuple(
                            StackContext(c, TensorSpecContext(s))
                            for s, c in self._variables[variable].uses
                        )
                    ),
                )
                for variable in sorted(variable_type_errors)
            ]
            raise VariableTypeError(ParallelContext(tuple(error_contexts)))

        return shape_checks

    def _match_dims(
        self, shape_check: _ShapeCheck
    ) -> Iterable[Tuple[ParsedDimensionSpec, _Shape, bool]]:
        """
        Match expected dimensions against actual dimensions.

        If some dimensions cannot be determined, the remaining dimensions are added to
        `_VariableState.waiting_for_varrank`.
        """
        # Determine rank of expected dimensions:

        # Length of the current shape, which we have not yet matched against a dimension spec:
        unallocated_len = shape_check.actual_end - shape_check.actual_begin

        # Length of dimension specs, `None` if it cannot be determined:
        allocated_sizes: List[Optional[int]] = []

        # Count of `None`s in `allocated_sizes`:
        n_allocated_none = 0

        # Mapping from variables, that we don't know the rank of, to whether they're used with
        # `broadcast`.
        unknown_len_variables: Dict[Optional[str], bool] = {}

        # The length of a shape is allowed to be short if the dimension, and all leading dimensions
        # are `broadcast`, so pre-compute where this applies.
        leading_broadcast = True
        leading_broadcasts = []
        for i in range(0, shape_check.expected_end):
            expected_dim = shape_check.expected.dims[i]
            leading_broadcast &= expected_dim.broadcastable
            leading_broadcasts.append(leading_broadcast)

        for i in range(shape_check.expected_begin, shape_check.expected_end)[::-1]:
            leading_broadcast = leading_broadcasts[i]
            expected_dim = shape_check.expected.dims[i]
            expected_name = expected_dim.variable_name

            rank: Optional[int] = None
            if expected_dim.variable_rank:
                if expected_name is not None:
                    observed_dims = self._variables[expected_name].observed_dims
                    assert observed_dims is not None
                    if observed_dims.known_rank and (observed_dims.sizes is not None):
                        rank = len(observed_dims.sizes)
            else:
                rank = 1

            if rank is not None and leading_broadcast:
                all_trailing_shapes_known = n_allocated_none == 0
                if all_trailing_shapes_known:
                    # We know where the dimension ends, because everything after this has known
                    # size, so we can infer the start from the `unallocated_len`.
                    rank = min(rank, unallocated_len)
                else:
                    # We don't know where this dimension starts, because of broadcasting, and we
                    # don't know where it ends, because something after this has an unknown size
                    # - all bets are off...
                    rank = None

            if rank is None:
                n_allocated_none += 1
                unknown_len_variables[expected_name] = leading_broadcast
            else:
                unallocated_len -= rank

            allocated_sizes.append(rank)

        allocated_sizes = allocated_sizes[::-1]  # This list was built in reverse order.

        # Determine length of variables with outknown length:

        # Number of variables with unknown length.
        n_unknown = len(unknown_len_variables)

        if n_unknown == 0:
            self._assert(unallocated_len == 0)
        else:
            self._assert(unallocated_len >= 0)
            if n_unknown == 1:
                (expected_name,) = unknown_len_variables
                if expected_name is not None:
                    broadcastable = unknown_len_variables[expected_name]
                    if n_allocated_none <= 1 or (not broadcastable):
                        self._assert(unallocated_len % n_allocated_none == 0)
                        unknown_len = unallocated_len // n_allocated_none
                        allocated_sizes = [unknown_len if s is None else s for s in allocated_sizes]

        dim_checks = []

        # Take as much as we can from the left side:

        allocated_i = 0
        while not shape_check.finished:
            expected_rank = allocated_sizes[allocated_i]
            if expected_rank is None:
                break

            dim_checks.append(
                (
                    shape_check.expected.dims[shape_check.expected_begin],
                    shape_check.actual[
                        shape_check.actual_begin : shape_check.actual_begin + expected_rank
                    ],
                    leading_broadcasts[shape_check.expected_begin],
                )
            )
            allocated_i += 1
            shape_check.actual_begin += expected_rank
            shape_check.expected_begin += 1

        # Take as much as we can from the right side:

        allocated_i = len(allocated_sizes) - 1
        while not shape_check.finished:
            expected_rank = allocated_sizes[allocated_i]
            if expected_rank is None:
                break

            dim_checks.append(
                (
                    shape_check.expected.dims[shape_check.expected_end - 1],
                    shape_check.actual[
                        shape_check.actual_end - expected_rank : shape_check.actual_end
                    ],
                    leading_broadcasts[shape_check.expected_begin],
                )
            )
            allocated_i -= 1
            shape_check.actual_end -= expected_rank
            shape_check.expected_end -= 1

        if not shape_check.finished:
            waiting_for: Set[str] = set(unknown_len_variables) - {None}  # type: ignore[assignment]
            for name in waiting_for:
                self._variables[name].waiting_for_varrank.add(shape_check)

        return dim_checks

    def _check_dim(
        self,
        expected: ParsedDimensionSpec,
        actual_dims: _Shape,
        shape_possibly_truncated: bool,
        shape_checks: List[_ShapeCheck],
    ) -> None:
        """
        Checks that ``actual_dim`` matches ``expected``.

        Newly learned information may enable the evaluation of deferred shape checks - any such will
        be added to ``shape_checks``.
        """
        if expected.variable_rank:
            expected_name = expected.variable_name
            if expected_name is None:
                # Anonymous dimension spec - we don't care about the actual values.
                return

            variable = self._variables[expected_name]
            observed_dims = variable.observed_dims
            assert observed_dims is not None

            self._assert(
                observed_dims.check_and_update(
                    actual_dims, expected.broadcastable, shape_possibly_truncated
                )
            )
            if observed_dims.known_rank and variable.waiting_for_varrank:
                shape_checks.extend(variable.waiting_for_varrank)
                variable.waiting_for_varrank.clear()

        else:
            if not actual_dims:
                # Broadcastable dimension was not matched against anything.
                assert shape_possibly_truncated
                return

            (actual_dim,) = actual_dims
            if actual_dim is None:
                # Acutal dimension is unknown - nothing to check.
                return

            if expected.constant is not None:
                observed_dim = _ObservedDim()
                assert observed_dim.check_and_update(expected.constant, broadcast=False)
            elif expected.variable_name is not None:
                maybe_observed_dim = self._variables[expected.variable_name].observed_dim
                assert maybe_observed_dim is not None
                observed_dim = maybe_observed_dim
            else:
                # Anonymous dimension - we don't care about the actual value.
                return
            self._assert(observed_dim.check_and_update(actual_dim, expected.broadcastable))

    def _assert(self, condition: bool) -> None:
        """
        Raise a nicely formatted :class:`ShapeMismatchError` if ``condition`` is not ``True``.
        """
        if not condition:
            contexts: List[ErrorContext] = []
            contexts.extend(self._additional_context)

            for shape, tensor_spec, context in self._observed_shapes:
                shape_error_context: ErrorContext = ShapeContext(tensor_spec.shape, shape)
                if tensor_spec.note is not None:
                    shape_error_context = ParallelContext(
                        (NoteContext(tensor_spec.note), shape_error_context)
                    )
                contexts.append(StackContext(context, shape_error_context))
            raise ShapeMismatchError(ParallelContext(tuple(contexts)))



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Code for configuring check_shapes."
"""
from contextlib import contextmanager
from enum import Enum
from typing import Iterator, Union

import tensorflow as tf


class ShapeCheckingState(Enum):
    """
    Different states of whether to actually check shapes.
    """

    ENABLED = "enabled"
    """
    Always check shapes.
    """

    EAGER_MODE_ONLY = "eager_mode_only"
    """
    Only check shapes if `tf.inside_function()` is `False`.
    """

    DISABLED = "disabled"
    """
    Never check shapes.
    """

    def __bool__(self) -> bool:
        """
        Return whether we currently should check shapes.
        """
        if self is ShapeCheckingState.ENABLED:
            return True
        elif self is ShapeCheckingState.EAGER_MODE_ONLY:
            # pylint: disable=no-member
            return not tf.inside_function()
        else:
            assert self is ShapeCheckingState.DISABLED, self
            return False


_enabled = ShapeCheckingState.EAGER_MODE_ONLY


class DocstringFormat(Enum):
    """
    Enumeration of supported formats of docstrings.
    """

    SPHINX = "sphinx"
    """
    Rewrite docstrings in the `Sphinx <https://www.sphinx-doc.org/en/master/>`_ format.
    """

    NONE = "none"
    """
    Do not rewrite docstrings.
    """


_docstring_format = DocstringFormat.SPHINX

_function_call_precompute_enabled = False


def set_enable_check_shapes(enabled: Union[ShapeCheckingState, str, bool]) -> None:
    """
    Set whether to enable :mod:`check_shapes`.

    Check shapes has a non-zero impact on performance. If this is unacceptable to you, you can
    use this function to disable it.

    Example:

    .. literalinclude:: /examples/test_check_shapes_examples.py
       :start-after: [disable__manual]
       :end-before: [disable__manual]
       :dedent:

    See also :func:`disable_check_shapes`.
    """
    global _enabled
    if enabled == True:  # pylint: disable=singleton-comparison
        _enabled = ShapeCheckingState.ENABLED
    elif enabled == False:  # pylint: disable=singleton-comparison
        _enabled = ShapeCheckingState.DISABLED
    elif isinstance(enabled, str):
        _enabled = ShapeCheckingState(enabled)
    else:
        assert isinstance(enabled, ShapeCheckingState), type(ShapeCheckingState)
        _enabled = enabled


def get_enable_check_shapes() -> ShapeCheckingState:
    """
    Get whether to enable :mod:`check_shapes`.
    """
    return _enabled


@contextmanager
def disable_check_shapes() -> Iterator[None]:
    """
    Context manager that temporarily disables shape checking.

    Example:

    .. literalinclude:: /examples/test_check_shapes_examples.py
       :start-after: [disable__context_manager]
       :end-before: [disable__context_manager]
       :dedent:
    """
    old_value = get_enable_check_shapes()
    set_enable_check_shapes(ShapeCheckingState.DISABLED)
    try:
        yield
    finally:
        set_enable_check_shapes(old_value)


def set_rewrite_docstrings(docstring_format: Union[DocstringFormat, str, None]) -> None:
    """
    Set how :mod:`check_shapes` should rewrite docstrings.

    See :class:`DocstringFormat` for valid choices.
    """
    global _docstring_format
    if docstring_format is None:
        _docstring_format = DocstringFormat.NONE
    elif isinstance(docstring_format, str):
        _docstring_format = DocstringFormat(docstring_format)
    else:
        assert isinstance(docstring_format, DocstringFormat), type(docstring_format)
        _docstring_format = docstring_format


def get_rewrite_docstrings() -> DocstringFormat:
    """
    Get how :mod:`check_shapes` should rewrite docstrings.
    """
    return _docstring_format


def set_enable_function_call_precompute(enabled: bool) -> None:
    """
    Set whether to precompute function call path and line numbers for debugging.

    This is disabled by default, because it is (relatively) slow. Enabling this can give better
    error messages.

    Example:

    .. literalinclude:: /examples/test_check_shapes_examples.py
       :start-after: [disable_function_call_precompute]
       :end-before: [disable_function_call_precompute]
       :dedent:
    """
    global _function_call_precompute_enabled
    _function_call_precompute_enabled = enabled


def get_enable_function_call_precompute() -> bool:
    """
    Get whether to precompute function call path and line numbers for debugging.
    """
    return _function_call_precompute_enabled



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Decorator for checking the shapes of function using tf Tensors.
"""
import inspect
from functools import update_wrapper
from typing import Any, Callable, Sequence, cast

import tensorflow as tf

from .accessors import set_check_shapes
from .argument_ref import RESULT_TOKEN
from .base_types import C
from .checker import ShapeChecker
from .checker_context import set_shape_checker
from .config import get_enable_check_shapes
from .error_contexts import (
    ConditionContext,
    FunctionCallContext,
    FunctionDefinitionContext,
    NoteContext,
    ParallelContext,
    StackContext,
)
from .parser import parse_and_rewrite_docstring, parse_function_spec
from .specs import ParsedArgumentSpec


def null_check_shapes(func: C) -> C:
    """
    Annotates the given function so that it looks like it has shape checks, but without actually
    checking anything.

    This is necessary not to break ``@inherit_check_shapes`` when shape checking is disabled.
    """
    set_check_shapes(func, null_check_shapes)
    return func


def check_shapes(*specs: str, tf_decorator: bool = False) -> Callable[[C], C]:
    """
    Decorator that checks the shapes of tensor arguments.

    Example:

    .. literalinclude:: /examples/test_check_shapes_examples.py
       :start-after: [basic]
       :end-before: [basic]
       :dedent:

    :param specs: Specification of arguments to check. See: `Check specification`_.
    :param tf_decorator: Whether to wrap the shape check with
        ``tf.compat.v1.flags.tf_decorator.make_decorator``.
        Setting this `True` seems to solve some problems, particularly related to Keras models,
        but create some other problems, particularly related to branching on tensors.
    """
    if not get_enable_check_shapes():
        return null_check_shapes

    unbound_error_context = FunctionCallContext(check_shapes)

    func_spec = parse_function_spec(specs, unbound_error_context)

    pre_specs = [spec for spec in func_spec.arguments if not spec.argument_ref.is_result]
    post_specs = [spec for spec in func_spec.arguments if spec.argument_ref.is_result]
    note_specs = func_spec.notes

    def _check_shapes(func: C) -> C:
        bound_error_context = FunctionDefinitionContext(func)
        signature = inspect.signature(func)

        def wrapped(*args: Any, **kwargs: Any) -> Any:
            if not get_enable_check_shapes():
                return func(*args, **kwargs)

            try:
                bound_arguments = signature.bind(*args, **kwargs)
            except TypeError as e:
                # TypeError is raised if *args and **kwargs don't actually match the arguments of
                # `func`. In that case we just call `func` normally, which will also result in an
                # error, but an error with the error message the user is used to.
                try:
                    func(*args, **kwargs)
                except TypeError as e2:
                    raise TypeError(
                        "Error calling wrapped function (see above error)."
                        " If you believe your parameters actually are correct, the error can"
                        " sometimes be fixed by setting `tf_decorator=True` on your `@check_shapes`"
                        " decorator."
                    ) from e2
                raise AssertionError(
                    "The above line should fail so this line should never be reached."
                ) from e
            bound_arguments.apply_defaults()
            arg_map = bound_arguments.arguments

            checker = ShapeChecker()
            for note_spec in note_specs:
                checker.add_context(StackContext(bound_error_context, NoteContext(note_spec)))

            def _check_specs(specs: Sequence[ParsedArgumentSpec]) -> None:
                processed_specs = []

                for arg_spec in specs:
                    for arg_value, relative_arg_context in arg_spec.argument_ref.get(
                        arg_map, bound_error_context
                    ):
                        arg_context = StackContext(bound_error_context, relative_arg_context)

                        if arg_spec.condition is not None:
                            condition, condition_context = arg_spec.condition.get(
                                arg_map,
                                StackContext(arg_context, ConditionContext(arg_spec.condition)),
                            )
                            if not condition:
                                continue
                            arg_context = StackContext(
                                bound_error_context,
                                ParallelContext(
                                    (
                                        StackContext(
                                            relative_arg_context,
                                            StackContext(
                                                ConditionContext(arg_spec.condition),
                                                condition_context,
                                            ),
                                        ),
                                    )
                                ),
                            )

                        processed_specs.append((arg_value, arg_spec.tensor, arg_context))

                checker.check_shapes(processed_specs)

            _check_specs(pre_specs)

            with set_shape_checker(checker):
                result = func(*args, **kwargs)
            arg_map[RESULT_TOKEN] = result

            _check_specs(post_specs)

            return result

        # Make TensorFlow understand our decoration:
        if tf_decorator:
            tf.compat.v1.flags.tf_decorator.make_decorator(func, wrapped)

        update_wrapper(wrapped, func)
        set_check_shapes(wrapped, _check_shapes)
        wrapped.__doc__ = parse_and_rewrite_docstring(func.__doc__, func_spec, bound_error_context)
        return cast(C, wrapped)

    return _check_shapes



# Copyright 2022 The GPflow Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Definitions of commonly used types.
"""
from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, TypeVar, Union

import numpy as np

from .type_flags import GENERIC_NP_ARRAYS, NP_TYPE_CHECKING

C = TypeVar("C", bound=Callable[..., Any])

Dimension = Optional[int]
"""
The size of a single observed dimension.

Use `None` if the size of that dimension is unknown.
"""

Shape = Optional[Tuple[Dimension, ...]]
"""
The complete shape of an observed object.

Use `None` if the object has a shape, but the shape is unknown.

Raise an exception if objects of that type can never have a shape.
"""

if TYPE_CHECKING and (not NP_TYPE_CHECKING):  # pragma: no cover
    AnyNDArray = Any
else:
    if GENERIC_NP_ARRAYS:
        # It would be nice to use something more interesting than `Any` here, but it looks like
        # the infrastructure in the rest of the ecosystem isn't really set up for this
        # yet. Maybe when we get Python 3.11?
        AnyNDArray = np.ndarray[Any, Any]  # type: ignore[misc]
    else:
        AnyNDArray = Union[np.ndarray]  # type: ignore[misc]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['check_shapes']

package_data = \
{'': ['*']}

install_requires = \
['lark>=1.1.0,<2.0.0',
 'tensorflow-probability>=0.12.0',
 'tensorflow>=2.4.0,<3.0.0']

setup_kwargs = {
    'name': 'check-shapes',
    'version': '0.1.0',
    'description': 'A library for annotating and checking the shapes of tensors.',
    'long_description': None,
    'author': 'Jesper Nielsen',
    'author_email': 'jespernielsen1982+check_shapes@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gpflow.github.io/check_shapes',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
